
import json
from SohuPrompt.data_utils import InputFeatures
import os
import torch
from torch import nn
from typing import *
from transformers import PreTrainedModel
from transformers.tokenization_utils import PreTrainedTokenizer
from SohuPrompt import Verbalizer
from SohuPrompt.prompts import One2oneVerbalizer, PtuningTemplate

class PTRTemplate(PtuningTemplate):
    """
    Args:
        model (:obj:`PreTrainedModel`): The pre-trained language model for the current prompt-learning task.
        tokenizer (:obj:`PreTrainedTokenizer`): A tokenizer to appoint the vocabulary and the tokenization strategy.
        text (:obj:`Optional[List[str]]`, optional): manual template format. Defaults to None.
        soft_token (:obj:`str`, optional): The special token for soft token. Default to ``<soft>``
        placeholder_mapping (:obj:`dict`): A place holder to represent the original input text. Default to ``{'<text_a>': 'text_a', '<text_b>': 'text_b'}``
    """
    def __init__(self,
                 model: PreTrainedModel,
                 tokenizer: PreTrainedTokenizer,
                 text:  Optional[str] = None,
                 placeholder_mapping: dict = {'<text_a>':'text_a', '<text_b>':'text_b'},
                ):
        super().__init__(model=model,
                         tokenizer=tokenizer,
                         prompt_encoder_type="mlp",
                         text=text,
                         placeholder_mapping=placeholder_mapping)


class PTRVerbalizer(Verbalizer):
    """
    In `PTR <https://arxiv.org/pdf/2105.11259.pdf>`_, each prompt has more than one ``<mask>`` tokens.
    Different ``<mask>`` tokens have different label words.
    The final label is predicted jointly by these label words using logic rules.

    Args:
        tokenizer (:obj:`PreTrainedTokenizer`): A tokenizer to appoint the vocabulary and the tokenization strategy.
        classes (:obj:`Sequence[str]`): A sequence of classes that need to be projected.
        label_words (:obj:`Union[Sequence[Sequence[str]], Mapping[str, Sequence[str]]]`, optional): The label words that are projected by the labels.
    """
    def __init__(self,
                 tokenizer: PreTrainedTokenizer,
                 classes: Sequence[str] = None,
                 num_classes: Optional[int] = None,
                 label_words: Optional[Union[Sequence[Sequence[str]], Mapping[str, Sequence[str]]]] = None,
                ):
        super().__init__(tokenizer = tokenizer, classes = classes, num_classes = num_classes)
        self.label_words = label_words

    def on_label_words_set(self):
        """
        Prepare One2oneVerbalizer for each `<mask>` separately
        """
        super().on_label_words_set()

        self.num_masks = len(self.label_words[0])
        for words in self.label_words:
            if len(words) != self.num_masks:
                raise ValueError("number of mask tokens for different classes are not consistent")
        self.sub_labels = [
            list(set([words[i] for words in self.label_words]))
            for i in range(self.num_masks)
        ] # [num_masks, label_size of the corresponding mask]

        self.verbalizers = nn.ModuleList([
            One2oneVerbalizer(tokenizer=self.tokenizer, label_words=labels, post_log_softmax = False)
            for labels in self.sub_labels
        ]) # [num_masks]

        self.label_mappings = nn.Parameter(torch.LongTensor([
            [labels.index(words[j]) for words in self.label_words]
            for j, labels in enumerate(self.sub_labels)
        ]), requires_grad=False) # [num_masks, label_size of the whole task]

    def process_logits(self,
                       logits: torch.Tensor, # [batch_size, num_masks, vocab_size]
                       batch: Union[Dict, InputFeatures],
                       **kwargs):
        """
        1) Process vocab logits of each `<mask>` into label logits of each `<mask>`

        2) Combine these logits into a single label logits of the whole task

        Args:
            logits (:obj:`torch.Tensor`): vocab logits of each `<mask>` (shape: `[batch_size, num_masks, vocab_size]`)

        Returns:
            :obj:`torch.Tensor`: logits (label logits of whole task (shape: `[batch_size, label_size of the whole task]`))
        """
        each_logits = [ # logits of each verbalizer
            self.verbalizers[i].process_logits(logits = logits[:, i, :], batch = batch, **kwargs)
            for i in range(self.num_masks)
        ] # num_masks * [batch_size, label_size of the corresponding mask]

        label_logits = [
            logits[:, self.label_mappings[j]]
            for j, logits in enumerate(each_logits)
        ]

        logsoftmax = nn.functional.log_softmax(sum(label_logits), dim=-1)

        if 'label' in batch: # TODO not an elegant solution
            each_logsoftmax = [ # (logits of each label) of each mask
                nn.functional.log_softmax(logits, dim=-1)[:, self.label_mappings[j]]
                for j, logits in enumerate(each_logits)
            ] # num_masks * [batch_size, label_size of the whole task]

            return logsoftmax + sum(each_logsoftmax) / len(each_logits) # [batch_size, label_size of the whole task]

        return logsoftmax



from .processor import *

class CMeEE_NER(CLSProcessor):
    """
    @inproceedings{hongying2020building,
    title={Building a Pediatric Medical Corpus: Word Segmentation and Named Entity Annotation},
    author={Hongying, Zan and Wenxin, Li and Kunli, Zhang and Yajuan, Ye and Baobao, Chang and Zhifang, Sui},
    booktitle={Workshop on Chinese Lexical Semantics},
    pages={652--664},
    year={2020},
    organization={Springer}
    }
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "dis", "sym", "dru", "equ", "pro", "bod", "ite", "mic", "dep",
            ],
            labels_mapped = [
                "疾病", "临床表现", "药物", "医疗设备", "医疗程序", "身体", "医学检验项目", "微生物类", "科室",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.json")
        examples = []
        with open(path, encoding='utf8') as f:
            for example_json in json.load(f):
                for span in example_json["entities"]:

                    example = InputExample(
                        meta = {
                            "context": example_json["text"],
                            "entity": span["entity"],
                            "options": self.labels_mapped,
                        },
                        label = self.get_label(span["type"])
                    )
                    examples.append(example)
        return examples
                
        

    def get_templates(self):
        return [
            '文本：{context} 问题：上文中，实体“{entity}”是什么类型的? {options}',
        ]


class Resume_NER(CLSProcessor):
    """
    @article{zhang2018chinese,  
    title={Chinese NER Using Lattice LSTM},  
    author={Yue Zhang and Jie Yang},  
    booktitle={Proceedings of the 56th Annual Meeting of the Association for Computational Linguistics (ACL)},
    year={2018}  
    }
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "NAME",  "CONT", "LOC", "RACE", "PRO", "EDU", "ORG", "TITLE",
            ],
            labels_mapped = [
                "人名", "国籍", "籍贯", "种族", "专业", "学位", "机构", "职称",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.jsonline")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                for span in example_json["span_list"]:
                    example = InputExample(
                        meta = {
                            "context": "".join(example_json["tokens"]),
                            "entity":  "".join(example_json["tokens"][span["start"]: span["end"]+1]),
                            "options": self.labels_mapped,
                        },
                        tgt_text = self.get_label(span["type"]),
                    )
                    examples.append(example)
                
        

    def get_templates(self):
        return [
            '文本：{context} 问题：上文中，实体“{entity}”是什么类型的? {options}',
        ]


class Weibo_NER(CLSProcessor):
    """
    @inproceedings{peng2015ner,
    title={Named Entity Recognition for Chinese Social Media with Jointly Trained Embeddings},
    author={Peng, Nanyun and Dredze, Mark},
    booktitle={Processings of the Conference on Empirical Methods in Natural Language Processing (EMNLP)},
    pages={548–-554},
    year={2015}, File={https://www.aclweb.org/anthology/D15-1064/}, }

    @inproceedings{peng2016improving,
    title={Improving named entity recognition for Chinese social media with word segmentation representation learning},
    author={Peng, Nanyun and Dredze, Mark},
    booktitle={Proceedings of the 54th Annual Meeting of the Association for Computational Linguistics (ACL)},
    volume={2},
    pages={149--155},
    year={2016}, File={https://www.aclweb.org/anthology/P16-2025/}, }

    @inproceedings{he-sun-2017-f,
    title = "{F}-Score Driven Max Margin Neural Network for Named Entity Recognition in {C}hinese Social Media",
    author = "He, Hangfeng  and
    Sun, Xu",
    booktitle = "Proceedings of the 15th Conference of the {E}uropean Chapter of the Association for Computational Linguistics: Volume 2, Short Papers",
    month = apr,
    year = "2017",
    address = "Valencia, Spain",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/E17-2113",
    pages = "713--718",
    }
    """
    def __init__(self):
        super().__init__(
            labels_origin = [ # TODO 区分 NAM 特指 NOM 泛指
                "PER.NAM",  "PER.NOM", "LOC.NAM", "LOC.NOM", "ORG.NAM", "ORG.NOM", "GPE.NAM", "GPE.NOM", 
            ],
            labels_mapped = [
                "人", "泛指人", "地点", "泛指地点", "机构", "泛指机构", "地理政治实体", "泛指地理政治实体", 
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.jsonline")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                for span in example_json["span_list"]:
                    example = InputExample(
                        meta = {
                            "context": "".join(example_json["tokens"]),
                            "entity": "".join(example_json["tokens"][span["start"]: span["end"]+1]),
                            "options": self.labels_mapped,
                        },
                        tgt_text = self.get_label(span["type"]),
                    )
                    examples.append(example)
                
        

    def get_templates(self):
        return [
            '文本：{context} 问题：上文中，实体“{entity}”是什么类型的? {options}',
        ]


class DH_MSRA(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "PER", "LOC", "ORG"
            ],
            labels_mapped = [
                "人", "地点", "机构"
            ]
        )

    def get_examples(self, data_dir, split):
        if split != 'train': raise ValueError
        path = os.path.join(data_dir, f"{split}.txt")
        
        with open(path, encoding='utf8') as f:
            xs, ys = [], []
            for line in f:
                l = line.split()
                if len(l)==0:
                    i = 0
                    while i < len(xs):
                        if ys[i][0] == 'B':
                            j = i + 1
                            while j < len(xs):
                                if ys[j][0] == 'O': break
                                j = j + 1

                            example = InputExample(
                                meta = {
                                    "context": "".join(xs),
                                    "entity": "".join(xs[i:j]),
                                    "options": self.labels_mapped,
                                },
                                tgt_text = self.get_label(ys[i][2:]),
                            )
                            examples.append(example)
                            
                            i = j
                        else:
                            i = i + 1

                    xs, ys = [], []
                else:
                    xs.append(l[0])
                    ys.append(l[1])

                
        

    def get_templates(self):
        return [
            '文本：{context} 问题：上文中，实体“{entity}”是什么类型的? {options}',
        ]



from transformers.tokenization_utils import PreTrainedTokenizer
from SohuPrompt.plms.utils import TokenizerWrapper
from typing import List, Dict, Optional
from collections import defaultdict
import numpy as np
from SohuPrompt.utils.logging import logger


class GLMTokenizerWrapper(TokenizerWrapper):
    r"""
    GLMTokenizer is a general language model from THUDM. GLM is a new model structure similar to GPT .
    Therefore it can only predict <mask> position
    at the end of the sentence. A prefix-style template like: 'A <mask> news : <text_a> <text_b> ' is
    not applicable in this situation.
    For the template where there is '<text_a>' or '<text_b>' after '<mask>', we raise an exception and terminate
    the program.
    For the template where there are template words after '<mask>', we ignore these template words.
    Moreover, it can only predict one '<mask>' position. All template that has multiple '<mask>' will
    give rise to an exception.
    """

    def __init__(
        self,
        max_seq_length: int,
        tokenizer: PreTrainedTokenizer,
        truncate_method: Optional[str] = "head",
        predict_eos_token: Optional[bool] = False,
        **kwargs
    ):
        super().__init__(
            max_seq_length=max_seq_length,
            tokenizer=tokenizer,
            truncate_method=truncate_method,
        )
        self.predict_eos = predict_eos_token

    @property
    def num_special_tokens_to_add(self):
        if not hasattr(self, "_num_specials"):
            self._num_specials = self.tokenizer.num_special_tokens_to_add()
        return self._num_specials

    def tokenize_one_example(self, wrapped_example, teacher_forcing):
        """# TODO doesn't consider the situation that input has two parts"""
        wrapped_example, others = wrapped_example
        mask_num = 0
        for i, example in enumerate(wrapped_example[::-1]):
            if mask_num==1:
                wrapped_example.insert(-i, {
                    "text":self.tokenizer.mask_token+self.tokenizer.bos_token,
                    "soft_token_ids":0,
                    "loss_ids":0,
                    "shortenable_ids":0
                })
                break
            if example['text'] == self.template_mask_token:
                mask_num+=1
        if teacher_forcing:
            tgt_text = others["tgt_text"]
            if isinstance(tgt_text, str):
                tgt_text = [tgt_text]
        
        if self.predict_eos:
            if not wrapped_example[-1]["text"].endswith(self.tokenizer.eop_token):
                wrapped_example.append(
                    {
                        "last": self.tokenizer.eop_token_id,
                        "shortenable_ids": 0,
                        "loss_ids": 1,
                        "soft_token_ids": 0
                    }
                )

        encoder_inputs = defaultdict(list)

        num_mask_token_used = 0
        mask_position = -1
        for piece_id, piece in enumerate(wrapped_example):
            if "last" in piece:
                encoder_inputs["input_ids"].append([piece['last']])
                for key in piece:
                    if key == "last":
                        continue
                    if key not in ["text"]:
                        encoder_inputs[key].append([piece[key]])

                continue
            if (
                piece["text"] == self.tokenizer.eop_token
                and self.predict_eos
                and wrapped_example[piece_id - 1]["loss_ids"] == 1
            ):  # eos after the mask also need to be pred
                piece["loss_ids"] = 1

            if piece["text"] == self.template_mask_token:
                if teacher_forcing:
                    piece["text"] = " " + tgt_text[num_mask_token_used] + " "
                else:
                    encoder_inputs["loss_ids"][-1][-1] = 1
                    break

            if piece["text"] in self.special_tokens_maps.keys():
                to_replace = self.special_tokens_maps[piece["text"]]
                if to_replace is not None:
                    piece["text"] = to_replace
                else:
                    raise KeyError(
                        "This tokenizer doesn't specify {} token.".format(piece["text"])
                    )

            if "soft_token_ids" in piece and piece["soft_token_ids"] != 0:
                encode_text = [
                    0
                ]  # can be replace by any token, since these token will use their own embeddings
            else:
                # if "mask_position" in piece:
                    # mask_position = piece_id
                encode_text = self.tokenizer.encode(piece["text"], add_special_tokens=False) 

            encoding_length = len(encode_text)

            encoder_inputs["input_ids"].append(encode_text)
            for key in piece:
                if key not in ["text"]:
                    encoder_inputs[key].append([piece[key]] * encoding_length)
        encoder_inputs = self.truncate(encoder_inputs=encoder_inputs, num_special_times=1)
        # encoder_inputs.pop('mask_position')
        # for key in encoder_inputs:
        #     if key == "input_ids":
        #         encoder_inputs[key][mask_position].extend([self.tokenizer.mask_token_id,self.tokenizer.bos_token_id])
        #     elif key == "soft_token_ids":
        #         encoder_inputs[key][mask_position].extend([0,0])
        #     elif key == "loss_ids":
        #         encoder_inputs[key][mask_position].extend([0,0])
        #     else:
        #         encoder_inputs[key][mask_position].extend([1,1])
        # delete shortenable ids
        encoder_inputs.pop("shortenable_ids")
        encoder_inputs = self.concate_parts(input_dict=encoder_inputs)
        # create special input ids
        encoder_inputs["attention_mask"] = [1] * len(encoder_inputs["input_ids"])
        if self.create_token_type_ids:
            encoder_inputs["token_type_ids"] = [0] * len(encoder_inputs["input_ids"])
        # pad to max length
        input_ids_len = len(encoder_inputs["input_ids"])
        encoder_inputs = self.padding(
            input_dict=encoder_inputs,
            max_len=self.max_seq_length,
            pad_id_for_inputs=self.tokenizer.pad_token_id,
        )
        encoder_inputs = {**encoder_inputs, "input_ids_len": input_ids_len}
        return encoder_inputs



# -*- coding: utf-8 -*-
import logging
import os
import datetime

logger = logging.getLogger()

def config_experiment_dir(config):
    r""" Automatic generate log directory for experiments.
    First generate the unique_string of one experiment, if the user
    didn't specify one, according
    to the user-defined keys logging.unique_string_keys.
    Then create the directory.
    """
    if not os.path.exists(config.logging.path_base):
        print(f"logging base directory `{config.logging.path_base}` not found, we help you creating one.")
        os.mkdir(config.logging.path_base)
        #raise NotADirectoryError(f"logging base directory `{config.logging.path_base}` not found, you should create one.")
    
    # generate unique string
    temp_strs = []
    if config.logging.unique_string is None:
        for item in config.logging.unique_string_keys:
            if item == "datetime":
                continue
            item = item.split(".") # split into sub items.
            subconfig = config
            for key in item:
                try:
                    subconfig = subconfig[key]
                except:
                    raise ValueError("The unique string_key is not a config option ")
            if not isinstance(subconfig, str):
                try:
                    subconfig = str(subconfig)
                except:
                    print("The value of subconfig key {} can't be converted to a string".format(".".join(item)))
                    continue
            
            subconfig = subconfig.split("/")[-1]
            temp_strs.append(subconfig)
        
        if 'datetime' in config.logging.unique_string_keys:
            if config.logging.datetime_format is None:
                config.logging.datetime_format = '%y%m%d%H%M%S'
            time_str = datetime.datetime.now().strftime(config.logging.datetime_format)
            temp_strs.append(time_str)
        config.logging.unique_string = "_".join(temp_strs)
    config.logging.path = os.path.join(config.logging.path_base, config.logging.unique_string)
    
    # create the log directory
    if os.path.exists(config.logging.path):
        if config.logging.overwrite:
            import shutil
            shutil.rmtree(config.logging.path)
            os.mkdir(config.logging.path)
        else:
            raise FileExistsError("Log dir {} exists and can't overwrite!")
    else:
        os.mkdir(config.logging.path)
    return config.logging.path

 
def init_logger(
    log_file,
    log_file_level=logging.NOTSET,
    log_level=logging.INFO,
):  
    if isinstance(log_file_level, str):
        log_file_level = getattr(logging, log_file_level)
    if isinstance(log_level, str):
        log_level = getattr(logging, log_level)
    log_format = logging.Formatter("[\033[032m%(asctime)s\033[0m %(levelname)s] %(module)s.%(funcName)s %(message)s")
    logger = logging.getLogger()
    logger.setLevel(log_level)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_format)
    logger.handlers = [console_handler]

    if log_file and log_file != '':
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(log_file_level)
        file_handler.setFormatter(log_format)
        logger.addHandler(file_handler)
    return logger
